"""CLI Support is here."""
import argparse
import asyncio
import inspect
import logging
import multiprocessing
import os
import re
import sys

from muffin import CONFIGURATION_ENVIRON_VARIABLE, __version__


PARAM_RE = re.compile(r'^\s+:param (\w+): (.+)$', re.M)


class Manager(object):

    """Provide simple interface to make application's commands.

    ::

        @app.manage.command
        def hello(name, upper=False):
            '''Say hello!

            :param name: an user name
            :param upper: to upper string

            '''
            message = 'Hello %s!' % name
            if upper:
                message = message.upper()
            print(message)

    ::

        $ muffin app hello Mike
        Hello Mike!

        $ muffin app hello Mike --upper
        HELLO MIKE!

    """

    def __init__(self, app):
        """Initialize the commands."""
        self.app = app
        self.parser = argparse.ArgumentParser(description="Manage %s" % app.name.capitalize())
        self.parsers = self.parser.add_subparsers(dest='subparser')
        self.handlers = dict()

        def shell_ctx():
            return {'app': app, 'run': lambda coro: app.loop.run_until_complete(coro)}

        app.cfg.setdefault('MANAGE_SHELL', shell_ctx)

        @self.command
        def shell(ipython: bool=True):
            """Run the application shell.

            :param ipython: Use IPython as shell
            """
            loop = asyncio.get_event_loop()
            app._set_loop(loop)
            app.freeze()
            loop.run_until_complete(app.startup())

            banner = '\nInteractive Muffin Shell\n'
            namespace = app.cfg.MANAGE_SHELL
            if callable(namespace):
                namespace = namespace()
            banner += "Loaded objects: %s" % list(namespace.keys())
            if ipython:
                try:
                    from IPython.terminal.embed import InteractiveShellEmbed
                    sh = InteractiveShellEmbed(banner1=banner)
                except ImportError:
                    pass
                else:
                    sh(local_ns=namespace)
                    return

            from code import interact
            interact(banner, local=namespace)

            loop.run_until_complete(app.cleanup())
            loop.run_until_complete(app.shutdown())

        workers = 1
        debug = app.cfg.DEBUG is not ... and app.cfg.DEBUG
        if not debug:
            workers = multiprocessing.cpu_count()

        @self.command
        def run(bind: str='127.0.0.1:5000', daemon: bool=False, pid: str=None,
                reload: bool=debug, timeout: int=30, name: str=self.app.name,
                worker_class: str='aiohttp.worker.GunicornWebWorker', workers: int=workers,
                log_file: str=None, access_logfile: str=self.app.cfg.ACCESS_LOG):
            """Run the application.

            :param bind: The socket to bind
            :param daemon: Daemonize the program
            :param log_file: The Error log file to write to
            :param name: A base to use with setproctitle for process naming
            :param pid: A filename to use for the PID file
            :param reload: Restart workers when code changes
            :param timeout: Workers silent for more than this many seconds are killed and restarted
            :param worker_class: The type of workers to use.
            :param workers: The number of worker processes for handling requests

            """
            from gunicorn.app.base import Application
            from gunicorn.util import import_app

            class MuffinServer(Application):

                def init(self, *args):
                    """Store config in env."""
                    os.environ[CONFIGURATION_ENVIRON_VARIABLE] = app.cfg.CONFIG or ''

                def load(self):
                    """Reload modules."""
                    if not app.uri:
                        return app

                    module, *_ = app.uri.split(':', 1)
                    if module in sys.modules:
                        sys.modules.pop(module)
                        paths = [p for p in sys.modules if p.startswith('%s.' % module)]
                        for path in paths:
                            sys.modules.pop(path)
                    return import_app(app.uri)

            server = MuffinServer('%(prog)s run [OPTIONS]')
            server.cfg.set('bind', bind)
            server.cfg.set('daemon', daemon)
            server.cfg.set('pidfile', pid)
            server.cfg.set('proc_name', name)
            server.cfg.set('reload', reload)
            server.cfg.set('timeout', timeout)
            server.cfg.set('worker_class', worker_class)
            if workers:
                server.cfg.set('workers', workers)

            if log_file:
                server.cfg.set('errorlog', log_file)
            if access_logfile:
                server.cfg.set('accesslog', access_logfile)

            server.run()

    def command(self, func):
        """Define CLI command."""
        header = '\n'.join([s for s in (func.__doc__ or '').split('\n')
                            if not s.strip().startswith(':')])
        parser = self.parsers.add_parser(func.__name__, description=header)
        args, vargs, kw, defs, kwargs, kwdefs, anns = inspect.getfullargspec(func)
        defs = defs or []
        kwargs_ = dict(zip(args[-len(defs):], defs))
        docs = dict(PARAM_RE.findall(func.__doc__ or ""))

        for name in args:
            argname = name.replace('_', '-').lower()
            arghelp = docs.get(name, '')

            if name not in kwargs_:
                parser.add_argument(argname, help=arghelp)
                continue

            value = kwargs_[name]

            if isinstance(value, bool):
                if value:
                    parser.add_argument("--no-" + argname, dest=name, action="store_false",
                                        help="Disable %s" % (arghelp or name).lower())
                else:
                    parser.add_argument("--" + argname, dest=name, action="store_true",
                                        help="Enable %s" % (arghelp or name).lower())
                continue

            if isinstance(value, list):
                parser.add_argument("--" + argname, action="append",
                                    default=value, help=arghelp)
                continue

            parser.add_argument("--" + argname, type=anns.get(name, type(value)),
                                default=value, help=arghelp + ' [%s]' % repr(value))

        self.handlers[func.__name__] = func
        func.parser = parser
        return func

    def shell(self, func):
        """Set shell context function."""
        self.app.cfg.MANAGE_SHELL = func

    def __call__(self, *args, prog=False):
        """Parse arguments and run handler."""
        if prog:
            self.parser.prog = prog
        if not args:
            args = sys.argv[1:]
        args_, _ = self.parser.parse_known_args(args)
        kwargs = dict(args_._get_kwargs())

        handler = self.handlers.get(kwargs.pop('subparser'))
        if not handler:
            self.parser.print_help()
            sys.exit(1)

        try:
            res = handler(**kwargs)
            if asyncio.iscoroutine(res):
                loop = asyncio.get_event_loop()
                loop.run_until_complete(res)

            sys.exit(0)
        except Exception as e:
            sys.exit(e)


def run():
    """CLI endpoint."""
    sys.path.insert(0, os.getcwd())
    logging.basicConfig(level=logging.INFO, handlers=[logging.StreamHandler()])

    parser = argparse.ArgumentParser(description="Manage Application", add_help=False)
    parser.add_argument('app', metavar='app',
                        type=str, help='Application module path')
    parser.add_argument('--config', type=str, help='Path to configuration.')
    parser.add_argument('--version', action="version", version=__version__)

    args_, subargs_ = parser.parse_known_args(sys.argv[1:])
    if args_.config:
        os.environ[CONFIGURATION_ENVIRON_VARIABLE] = args_.config

    from gunicorn.util import import_app

    app_uri = args_.app
    if ':' not in app_uri:
        app_uri += ':app'
    try:
        app = import_app(app_uri)
        app.uri = app_uri
        app.logger.info('Application is loaded: %s' % app.name)
    except Exception as exc:
        logging.exception(exc)
        raise sys.exit(1)

    app.manage(*subargs_, prog='muffin %s' % args_.app)

# pylama:ignore=C901,W0612,W0703,W0212



"""Base handler class."""
import functools
import inspect
from asyncio import coroutine, iscoroutine, iscoroutinefunction

import ujson as json
from aiohttp.hdrs import METH_ANY, METH_ALL
from aiohttp.web import StreamResponse, HTTPMethodNotAllowed, Response

from muffin.urls import routes_register
from muffin.utils import to_coroutine


ROUTE_PARAMS_ATTR = '_route_params'


def register(*paths, methods=None, name=None, handler=None):
    """Mark Handler.method to aiohttp handler.

    It uses when registration of the handler with application is postponed.

    ::
        class AwesomeHandler(Handler):

            def get(self, request):
                return "I'm awesome!"

            @register('/awesome/best')
            def best(self, request):
                return "I'm best!"

    """
    def wrapper(method):
        """Store route params into method."""
        method = to_coroutine(method)
        setattr(method, ROUTE_PARAMS_ATTR, (paths, methods, name))
        if handler and not hasattr(handler, method.__name__):
            setattr(handler, method.__name__, method)
        return method
    return wrapper


class HandlerMeta(type):

    """Prepare handlers."""

    _coroutines = set(m.lower() for m in METH_ALL)

    def __new__(mcs, name, bases, params):
        """Prepare a Handler Class.

        Ensure that the Handler class has a name.
        Ensure that required methods are coroutines.
        Fix the Handler params.
        """
        # Set name
        params['name'] = params.get('name', name.lower())

        # Define new coroutines
        for fname, method in params.items():
            if iscoroutinefunction(method):
                mcs._coroutines.add(fname)

        cls = super().__new__(mcs, name, bases, params)

        # Ensure that the class methods are exist and iterable
        if not cls.methods:
            cls.methods = set(method for method in METH_ALL if method.lower() in cls.__dict__)

        elif isinstance(cls.methods, str):
            cls.methods = [cls.methods]

        cls.methods = [method.upper() for method in cls.methods]

        # Ensure that coroutine methods is coroutines
        for name in mcs._coroutines:
            method = getattr(cls, name, None)
            if not method:
                continue
            setattr(cls, name, to_coroutine(method))

        return cls


class Handler(object, metaclass=HandlerMeta):

    """Handle request."""

    app = None
    name = None
    methods = None

    @classmethod
    def from_view(cls, view, *methods, name=None):
        """Create a handler class from function or coroutine."""
        docs = getattr(view, '__doc__', None)
        view = to_coroutine(view)

        if METH_ANY in methods:
            methods = METH_ALL

        def proxy(self, *args, **kwargs):
            return view(*args, **kwargs)

        params = {m.lower(): proxy for m in methods}
        params['methods'] = methods
        if docs:
            params['__doc__'] = docs

        return type(name or view.__name__, (cls,), params)

    @classmethod
    def bind(cls, app, *paths, methods=None, name=None, router=None, view=None):
        """Bind to the given application."""
        cls.app = app
        if cls.app is not None:
            for _, m in inspect.getmembers(cls, predicate=inspect.isfunction):
                if not hasattr(m, ROUTE_PARAMS_ATTR):
                    continue
                paths_, methods_, name_ = getattr(m, ROUTE_PARAMS_ATTR)
                name_ = name_ or ("%s.%s" % (cls.name, m.__name__))
                delattr(m, ROUTE_PARAMS_ATTR)
                cls.app.register(*paths_, methods=methods_, name=name_, handler=cls)(m)

        @coroutine
        @functools.wraps(cls)
        def handler(request):
            return cls().dispatch(request, view=view)

        if not paths:
            paths = ["/%s" % cls.__name__]

        return routes_register(
            app, handler, *paths, methods=methods, router=router, name=name or cls.name)

    @classmethod
    def register(cls, *args, **kwargs):
        """Register view to handler."""
        if cls.app is None:
            return register(*args, handler=cls, **kwargs)
        return cls.app.register(*args, handler=cls, **kwargs)

    async def dispatch(self, request, view=None, **kwargs):
        """Dispatch request."""
        if view is None and request.method not in self.methods:
            raise HTTPMethodNotAllowed(request.method, self.methods)

        method = getattr(self, view or request.method.lower())
        response = await method(request, **kwargs)
        return await self.make_response(request, response)

    __iter__ = dispatch

    async def make_response(self, request, response):
        """Convert a handler result to web response."""
        while iscoroutine(response):
            response = await response

        if isinstance(response, StreamResponse):
            return response

        if isinstance(response, str):
            return Response(text=response, content_type='text/html')

        if isinstance(response, bytes):
            return Response(body=response, content_type='text/html')

        return Response(text=json.dumps(response), content_type='application/json')

    @staticmethod
    def parse(request):
        """Return a coroutine which parses data from request depends on content-type.

        Usage: ::

            def post(self, request):
                data = yield from self.parse(request)
                # ...
        """
        if request.content_type in ('application/x-www-form-urlencoded', 'multipart/form-data'):
            return request.post()

        if request.content_type == 'application/json':
            return request.json()

        return request.text()



import logging
import os
import asyncio

import pytest

from aiohttp.pytest_plugin import pytest_addoption as aiohttp_pytest_addoption
from aiohttp.pytest_plugin import *  # noqa


def pytest_addoption(parser):
    """ Append pytest options for testing Muffin apps. """
    parser.addini('muffin_app', 'Set path to muffin application')
    parser.addoption('--muffin-app', dest='muffin_app', help='Set to muffin application')

    parser.addini('muffin_config', 'Set module path to muffin configuration')
    parser.addoption('--muffin-config', dest='muffin_config',
                     help='Set module path to muffin configuration')

    aiohttp_pytest_addoption(parser)


def pytest_load_initial_conftests(early_config, parser, args):
    """ Prepare to loading Muffin application. """
    from muffin import CONFIGURATION_ENVIRON_VARIABLE

    options = parser.parse_known_args(args)

    # Initialize configuration
    config = options.muffin_config or early_config.getini('muffin_config')
    if config:
        os.environ[CONFIGURATION_ENVIRON_VARIABLE] = config

    # Initialize application
    app_ = options.muffin_app or early_config.getini('muffin_app')
    early_config.app = app_


@pytest.fixture(scope='session')
def app(pytestconfig, request):
    """ Provide an example application. """
    from muffin.utils import to_coroutine
    from gunicorn import util

    if pytestconfig.app:
        app = util.import_app(pytestconfig.app)

        loop = asyncio.get_event_loop()
        for plugin in app.ps.values():
            if not hasattr(plugin, 'conftest'):
                continue
            loop.run_until_complete(to_coroutine(plugin.conftest)())

        return app

    logging.warn(
        'Improperly configured. Please set ``muffin_app`` in your pytest config. '
        'Or use ``--muffin-app`` command option.')

    return None


@pytest.fixture
async def client(app, aiohttp_client, loop):
    """Dirty hack for aiohttp tests."""
    app._loop = loop
    return await aiohttp_client(app)


#  pylama:ignore=W0212,W0621



import hashlib
import hmac
import importlib
import pkgutil
import random
import sys
import threading
from asyncio import coroutine, iscoroutinefunction, Task, get_event_loop


SALT_CHARS = 'bcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'


class MuffinException(Exception):

    """ Implement a Muffin's exception. """

    pass


def to_coroutine(func):
    """ Ensure that the function is coroutine.

    If not convert to coroutine.

    """
    if not iscoroutinefunction(func):
        func = coroutine(func)
    return func


def create_signature(secret, value, digestmod='sha256', encoding='utf-8'):
    """ Create HMAC Signature from secret for value. """
    if isinstance(secret, str):
        secret = secret.encode(encoding)

    if isinstance(value, str):
        value = value.encode(encoding)

    if isinstance(digestmod, str):
        digestmod = getattr(hashlib, digestmod, hashlib.sha1)

    hm = hmac.new(secret, digestmod=digestmod)
    hm.update(value)
    return hm.hexdigest()


def check_signature(signature, *args, **kwargs):
    """ Check for the signature is correct. """
    return hmac.compare_digest(signature, create_signature(*args, **kwargs))


def generate_password_hash(password, digestmod='sha256', salt_length=8):
    """ Hash a password with given method and salt length. """

    salt = ''.join(random.sample(SALT_CHARS, salt_length))
    signature = create_signature(salt, password, digestmod=digestmod)
    return '$'.join((digestmod, salt, signature))


def check_password_hash(password, pwhash):
    if pwhash.count('$') < 2:
        return False
    digestmod, salt, signature = pwhash.split('$', 2)
    return check_signature(signature, salt, password, digestmod=digestmod)


class Struct(dict):

    """ `Attribute` dictionary. Use attributes as keys. """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError("Attribute '%s' doesn't exists. " % name)

    def __setattr__(self, name, value):
        self[name] = value


class LStruct(Struct):

    """ Frosen structure. Used as application/plugins settings.

    Going to be immutable after application is started.

    """

    def __init__(self, *args, **kwargs):
        object.__setattr__(self, 'frozen', False)
        super().__init__(*args, **kwargs)

    def freeze(self):
        object.__setattr__(self, 'frozen', True)

    def __setitem__(self, name, value):
        if self.frozen:
            raise RuntimeError('`%s` is frozen.' % type(self))
        super().__setitem__(name, value)


class local_storage:
    pass


class local:

    """ coroutine.local storage is simular to python's threading.local.

    Usage: ::

        local = muffin.local(loop)
        local.value = 42

    """

    __slots__ = '_loop',
    __loops__ = {}

    def __new__(cls, loop=None):
        """ The local is singleton per loop. """
        key = "%s%s" % (cls, id(loop))
        if key not in cls.__loops__:
            cls.__loops__[key] = object.__new__(cls)
        return cls.__loops__[key]

    def __init__(self, loop=None):
        """ Bind loop to self. """
        object.__setattr__(self, '_loop', loop or get_event_loop())

    def __getattribute__(self, name):
        """ Get attribute from current task's space. """
        if name in ('__setattr__', '__getattr__', '__delattr__', '_loop', '__curtask__'):
            return object.__getattribute__(self, name)
        return getattr(self.__curtask__, name)

    def __setattr__(self, name, value):
        """ Set attribute to current task's space. """
        self.__curtask__.__dict__[name] = value

    def __delattr__(self, name):
        delattr(self.__curtask__, name)

    @property
    def __curtask__(self):
        """ Create namespace in current task. """
        task = Task.current_task(loop=self._loop)
        if not task:
            raise RuntimeError('No task is currently running')

        if not hasattr(task, '_locals'):
            task._locals = local_storage()
        return task._locals


tlocals = threading.local()


class slocal(local):

    """ Fail silently to threading.local if curent loop is not running. """

    __loops__ = {}

    def __getattribute__(self, name):
        """ Get attribute from current task's space. """
        if name in ('__setattr__', '__getattr__', '__delattr__', '_loop', '__curtask__'):
            return object.__getattribute__(self, name)

        if self._loop.is_running():
            return getattr(self.__curtask__, name)
        return getattr(tlocals, name)

    def __setattr__(self, name, value):
        """ Set attribute to current task's space. """
        if self._loop.is_running():
            super().__setattr__(name, value)
        else:
            setattr(tlocals, name, value)


def import_submodules(package_name, *submodules):
    """Import all submodules by package name."""
    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + '.' + name)
        for _, name, _ in pkgutil.walk_packages(package.__path__)
        if not submodules or name in submodules
    }

#  pylama:ignore=W0212



"""Implement Muffin Application."""
import logging.config
import os
from importlib import import_module
from inspect import isfunction, isclass, ismethod
import types

from aiohttp import web, log
from aiohttp.hdrs import METH_ANY
from cached_property import cached_property

from . import CONFIGURATION_ENVIRON_VARIABLE
from .handler import Handler
from .manage import Manager
from .utils import LStruct, to_coroutine
from .urls import SafeStaticResource


class MuffinException(Exception):

    """Exception class for Muffin errors."""

    pass


class BaseApplication(web.Application):

    def register(self, *paths, methods=None, name=None, handler=None):
        """Register function/coroutine/muffin.Handler with the application.

        Usage example:

        .. code-block:: python

            @app.register('/hello')
            def hello(request):
                return 'Hello World!'


        """
        if isinstance(methods, str):
            methods = [methods]

        def wrapper(view):
            if handler is None:

                handler_ = view
                methods_ = methods or [METH_ANY]

                if isfunction(handler_) or ismethod(handler_):
                    handler_ = Handler.from_view(view, *methods_, name=name)

                handler_.bind(self, *paths, methods=methods_, name=name)

            else:

                view_name = view.__name__
                if not hasattr(handler, view_name):
                    setattr(handler, view_name, to_coroutine(view))
                name_ = name or view_name
                handler.bind(self, *paths, methods=methods, name=name_, view=view_name)

            return view

        # Support for @app.register(func)
        if len(paths) == 1 and callable(paths[0]):
            view = paths[0]

            if isclass(view) and issubclass(view, BaseException):
                return wrapper

            paths = []
            return wrapper(view)

        return wrapper


class Application(BaseApplication):

    """Upgrade Aiohttp Application."""

    # Default application options
    defaults = {

        # Path to configuration module
        'CONFIG': None,

        # Enable debug mode (optional)
        'DEBUG': ...,

        # Plugins to install
        'PLUGINS': [],

        # Logging options
        'ACCESS_LOG': '-',  # File path to access log, - to stderr
        'ACCESS_LOG_FORMAT': '%a %l %u %t "%r" %s %b "%{Referrer}i" "%{User-Agent}i"',
        'LOG_LEVEL': 'WARNING',
        'LOG_FORMAT': '%(asctime)s [%(process)d] [%(levelname)s] %(message)s',
        'LOG_DATE_FORMAT': '[%Y-%m-%d %H:%M:%S]',

        # Setup static files in development
        'STATIC_PREFIX': '/static',
        'STATIC_FOLDERS': [],
    }

    uri = None

    def __init__(self, name, *, logger=log.web_logger, middlewares=(), handler_args=None,
                 client_max_size=1024**2, debug=..., **OPTIONS):
        """Initialize the application."""
        super(Application, self).__init__(
            logger=logger, middlewares=middlewares, handler_args=handler_args,
            client_max_size=client_max_size, debug=debug)

        self.name = name

        # Overide options
        self.defaults['CONFIG'] = OPTIONS.pop('CONFIG', self.defaults['CONFIG'])
        self.cfg.update(OPTIONS)
        if self.cfg.DEBUG is not ...:
            self._debug = self.cfg.DEBUG

        # Setup logging
        if not self.logger.handlers:
            ch = logging.StreamHandler()
            ch.setFormatter(logging.Formatter(self.cfg.LOG_FORMAT, self.cfg.LOG_DATE_FORMAT))
            self.logger.addHandler(ch)
            self.logger.setLevel(self.cfg.LOG_LEVEL)
            self.logger.propagate = False

        LOGGING_CFG = self.cfg.get('LOGGING')
        if LOGGING_CFG and isinstance(LOGGING_CFG, dict):
            logging.config.dictConfig(LOGGING_CFG)

        # Setup CLI
        self.manage = Manager(self)

        self._error_handlers = {}

        # Setup static files
        if isinstance(self.cfg.STATIC_FOLDERS, str):
            self.cfg.STATIC_FOLDERS = [self.cfg.STATIC_FOLDERS]

        elif not isinstance(self.cfg.STATIC_FOLDERS, list):
            self.cfg.STATIC_FOLDERS = list(self.cfg.STATIC_FOLDERS)

        # Setup plugins
        self.plugins = self.ps = LStruct()
        for plugin in self.cfg.PLUGINS:
            try:
                self.install(plugin)
            except Exception as exc:  # noqa
                self.logger.error('Plugin is invalid: %s', plugin)
                self.logger.exception(exc)

    def __repr__(self):
        """Human readable representation."""
        return "<Application: %s>" % self.name

    @cached_property
    def cfg(self):
        """Load the application configuration.

        This method loads configuration from python module.
        """
        config = LStruct(self.defaults)
        module = config['CONFIG'] = os.environ.get(
            CONFIGURATION_ENVIRON_VARIABLE, config['CONFIG'])

        if module:
            try:
                module = import_module(module)
                config.update({
                    name: getattr(module, name) for name in dir(module)
                    if name == name.upper() and not name.startswith('_')
                })

            except ImportError as exc:
                config.CONFIG = None
                message = "Error importing %s: %s" % (module, exc)
                self.register_on_start(lambda app: app.logger.error(message))

        return config

    def install(self, plugin, name=None, **opts):
        """Install plugin to the application."""
        source = plugin

        if isinstance(plugin, str):
            module, _, attr = plugin.partition(':')
            module = import_module(module)
            plugin = getattr(module, attr or 'Plugin', None)

        if isinstance(plugin, types.ModuleType):
            plugin = getattr(module, 'Plugin', None)

        if plugin is None:
            raise MuffinException('Plugin is not found %r' % source)

        name = name or plugin.name
        if name in self.ps:
            raise MuffinException('Plugin with name `%s` is already intalled.' % name)

        if isinstance(plugin, type):
            plugin = plugin(**opts)

        if hasattr(plugin, 'setup'):
            plugin.setup(self)

        if hasattr(plugin, 'middleware') and plugin.middleware not in self.middlewares:
            self.middlewares.append(plugin.middleware)

        if hasattr(plugin, 'startup'):
            self.on_startup.append(plugin.startup)

        if hasattr(plugin, 'cleanup'):
            self.on_cleanup.append(plugin.cleanup)

        # Save plugin links
        self.ps[name] = plugin

        return plugin

    async def startup(self):
        """Start the application.

        Support for start-callbacks and lock the application's configuration and plugins.
        """
        if self.frozen:
            return False

        if self._error_handlers:
            self.middlewares.append(_exc_middleware_factory(self))

        # Register static paths
        for path in self.cfg.STATIC_FOLDERS:
            self.router.register_resource(SafeStaticResource(self.cfg.STATIC_PREFIX, path))

        await super(Application, self).startup()

    def pre_freeze(self) -> None:
        # Lock the application's settings and plugin's registry after start
        self.cfg.freeze()
        self.ps.freeze()

        # Freeze plugin's
        for plugin in self.ps.values():
            plugin.freeze()

        return super(Application, self).pre_freeze()


def _exc_middleware_factory(app):
    """Handle exceptions.

    Route exceptions to handlers if they are registered in application.
    """

    @web.middleware
    async def middleware(request, handler):
        try:
            return await handler(request)
        except Exception as exc:
            for cls in type(exc).mro():
                if cls in app._error_handlers:
                    request.exception = exc
                    response = await app._error_handlers[cls](request)
                    return response
            raise

    return middleware


def run():
    """Run Gunicorn Application."""
    from .worker import GunicornApp

    GunicornApp("%(prog)s [OPTIONS] [APP]").run()


if __name__ == '__main__':
    run()

#  pylama:ignore=W0212,W1504
